@@ -75,6 +75,9 @@ static const struct command_info game_cmds[] =
 	{ CMD_GO_DOWN, "go down stairs", do_cmd_go_down, false, true, 0 },
 	{ CMD_WALK, "walk", do_cmd_walk, true, true, 0 },
 	{ CMD_RUN, "run", do_cmd_run, true, true, 0 },
+	{ CMD_EXPLORE, "explore", do_cmd_explore, false, true, 0 },
+	{ CMD_NAVIGATE_UP, "navigate up", do_cmd_navigate_up, false, true, 0 },
+	{ CMD_NAVIGATE_DOWN, "navigate down", do_cmd_navigate_down, false, true, 0 },
 	{ CMD_JUMP, "jump", do_cmd_jump, false, true, 0 },
 	{ CMD_OPEN, "open", do_cmd_open, true, true, 99 },
 	{ CMD_CLOSE, "close", do_cmd_close, true, true, 99 },
@@ -95,11 +98,11 @@ static const struct command_info game_cmds[] =
 	{ CMD_AUTOINSCRIBE, "autoinscribe", do_cmd_autoinscribe, false, false, 0 },
 	{ CMD_EAT, "eat", do_cmd_eat_food, false, true, 0 },
 	{ CMD_QUAFF, "quaff", do_cmd_quaff_potion, false, true, 0 },
-	{ CMD_USE_ROD, "zap", do_cmd_zap_rod, false, true, 0 },
-	{ CMD_USE_STAFF, "use", do_cmd_use_staff, false, true, 0 },
-	{ CMD_USE_WAND, "aim", do_cmd_aim_wand, false, true, 0 },
+	{ CMD_USE_ROD, "zap", do_cmd_zap_rod, true, true, 99 },
+	{ CMD_USE_STAFF, "use", do_cmd_use_staff, true, true, 99 },
+	{ CMD_USE_WAND, "aim", do_cmd_aim_wand, true, true, 99 },
 	{ CMD_READ_SCROLL, "read", do_cmd_read_scroll, false, true, 0 },
-	{ CMD_ACTIVATE, "activate", do_cmd_activate, false, true, 0 },
+	{ CMD_ACTIVATE, "activate", do_cmd_activate, true, true, 99 },
 	{ CMD_REFILL, "refuel with", do_cmd_refill, false, true, 0 },
 	{ CMD_FIRE, "fire", do_cmd_fire, false, true, 0 },
 	{ CMD_THROW, "throw", do_cmd_throw, false, true, 0 },
@@ -110,8 +113,8 @@ static const struct command_info game_cmds[] =
 	{ CMD_STASH, "stash", do_cmd_stash, false, false, 0 },
 	{ CMD_BUY, "buy", do_cmd_buy, false, false, 0 },
 	{ CMD_RETRIEVE, "retrieve", do_cmd_retrieve, false, false, 0 },
-	{ CMD_USE, "use", do_cmd_use, false, true, 0 },
-	{ CMD_SUICIDE, "kill character", do_cmd_suicide, false, false, 0 },
+	{ CMD_USE, "use", do_cmd_use, true, true, 0 },
+	{ CMD_RETIRE, "retire character", do_cmd_retire, false, false, 0 },
 	{ CMD_HELP, "help", NULL, false, false, 0 },
 	{ CMD_REPEAT, "repeat", NULL, false, false, 0 },
 
@@ -171,6 +174,39 @@ static const struct command_info game_cmds[] =
 	{ CMD_WIZ_WIZARD_LIGHT, "wizard light the level", do_cmd_wiz_wizard_light, false, false, 0 },
 };
 
+/**
+ * Make a deep copy of a command and its arguments.
+ */
+void cmd_copy(struct command *dest, const struct command *src)
+{
+	int i;
+
+	*dest = *src;
+	/* String arguments require a deep copy. */
+	for (i = 0; i < CMD_MAX_ARGS; ++i) {
+		if (dest->arg[i].type == arg_STRING) {
+			dest->arg[i].data.string =
+				string_make(src->arg[i].data.string);
+		}
+	}
+}
+
+/**
+ * Release any resources for the command's arguments.
+ */
+void cmd_release(struct command *cmd)
+{
+	int i;
+
+	for (i = 0; i < CMD_MAX_ARGS; ++i) {
+		if (cmd->arg[i].type == arg_STRING) {
+			string_free((char*)(cmd->arg[i].data.string));
+			cmd->arg[i].name[0] = '\0';
+			cmd->arg[i].type = arg_NONE;
+		}
+	}
+}
+
 const char *cmd_verb(cmd_code cmd)
 {
 	size_t i;
@@ -208,42 +244,24 @@ static int cmd_head = 0;
 static int cmd_tail = 0;
 static struct command cmd_queue[CMD_QUEUE_SIZE];
 
+/*
+ * Remember last non-background command executed for use by CMD_REPEAT; either
+ * still in the queue at last_command_idx or copied out of the queue into
+ * last_command
+ */
+int last_command_idx = -1;
+static struct command last_command = {
+	.context = CTX_INIT,
+	.code = CMD_NULL,
+	.nrepeats = 0,
+	.background_command = 0,
+	.arg = { { 0 } }
+};
+
 static bool repeat_prev_allowed = false;
 static bool repeating = false;
 
 
-/**
- * Locate and return the index of the first previous command that is an
- * appropriate target for CMD_REPEAT.  If no such command is available,
- * return a negative value.
- */
-static int cmdq_find_repeat_target_index(void)
-{
-	int target = cmd_head - 1;
-
-	while (1) {
-		if (target < 0) target = CMD_QUEUE_SIZE - 1;
-
-		if (target == cmd_tail) {
-			break;
-		}
-		/*
-		 * Only repeat a command that has not been marked
-		 * as a background command (i.e. it is not a side
-		 * effect of something else).
-		 */
-		if (!cmd_queue[target].is_background_command) {
-			if (cmd_queue[target].code != CMD_NULL) {
-				return target;
-			}
-			break;
-		}
-		/* Keep looking. */
-		--target;
-	}
-	return -1;
-}
-
 struct command *cmdq_peek(void)
 {
 	return &cmd_queue[prev_cmd_idx(cmd_head)];
@@ -252,6 +270,9 @@ struct command *cmdq_peek(void)
 
 /**
  * Insert the given command into the command queue.
+ *
+ * Makes a shallow copy of the command.  If the command has string arguments,
+ * the queue assumes ownership of those arguments.
  */
 errr cmdq_push_copy(struct command *cmd)
 {
@@ -261,20 +282,38 @@ errr cmdq_push_copy(struct command *cmd)
 
 	/* Insert command into queue. */
 	if (cmd->code != CMD_REPEAT) {
+		if (last_command_idx == cmd_head) {
+			/*
+			 * The last non-background command will be overwritten.
+			 * Copy out of the queue in case it is needed for
+			 * CMD_REPEAT.
+			 */
+			cmd_release(&last_command);
+			cmd_copy(&last_command, &cmd_queue[cmd_head]);
+			last_command_idx = -1;
+		}
+		cmd_release(&cmd_queue[cmd_head]);
 		cmd_queue[cmd_head] = *cmd;
 	} else if (!repeat_prev_allowed) {
 		return 1;
 	} else {
-		int idx_to_repeat = cmdq_find_repeat_target_index();
-
-		if (idx_to_repeat < 0) {
-			return 1;
-		}
 		/*
 		 * If we're repeating a command, we duplicate the previous
 		 * command in the next command "slot".
 		 */
-		cmd_queue[cmd_head] = cmd_queue[idx_to_repeat];
+		if (last_command_idx >= 0) {
+			assert(last_command_idx < CMD_QUEUE_SIZE);
+			if (last_command_idx != cmd_head) {
+				cmd_release(&cmd_queue[cmd_head]);
+				cmd_copy(&cmd_queue[cmd_head],
+					&cmd_queue[last_command_idx]);
+			}
+		} else if (last_command.code != CMD_NULL) {
+			cmd_release(&cmd_queue[cmd_head]);
+			cmd_copy(&cmd_queue[cmd_head], &last_command);
+		} else {
+			return 1;
+		}
 	}
 
 	/* Advance point in queue, wrapping around at the end */
@@ -319,11 +358,12 @@ static void process_command(cmd_context ctx, struct command *cmd)
 	/* Actually execute the command function */
 	if (game_cmds[idx].fn) {
 		/* Occasional attack instead for bloodlust-affected characters */
-		if (cmd->is_background_command) {
+		if (cmd->background_command > 1) {
 			/*
-			 * Background commands do not trigger bloodlust.  If
-			 * they can take energy, they also don't reset whether
-			 * the player's next command skips the bloodlust check.
+			 * Some background commands do not trigger bloodlust.
+			 * If they can take energy, they also don't reset
+			 * whether the player's next command skips the
+			 * bloodlust check.
 			 */
 			if (player->skip_cmd_coercion
 					&& game_cmds[idx].can_use_energy) {
@@ -375,6 +415,9 @@ bool cmdq_pop(cmd_context c)
 	}
 
 	/* Now process it */
+	if (!cmd->background_command) {
+		last_command_idx = prev_cmd_idx(cmd_tail);
+	}
 	process_command(c, cmd);
 	return true;
 }
@@ -389,7 +432,7 @@ errr cmdq_push_repeat(cmd_code c, int nrepeats)
 		.context = CTX_INIT,
 		.code = CMD_NULL,
 		.nrepeats = 0,
-		.is_background_command = false,
+		.background_command = 0,
 		.arg = { { 0 } }
 	};
 
@@ -428,6 +471,22 @@ void cmdq_flush(void)
 	cmd_tail = cmd_head;
 }
 
+/**
+ * Remove all commands from the queue and release any allocated resources.
+ */
+void cmdq_release(void)
+{
+	int i;
+
+	cmdq_flush();
+	for (i = 0; i < CMD_QUEUE_SIZE; ++i) {
+		cmd_release(cmd_queue + i);
+	}
+	cmd_release(&last_command);
+	last_command.code = CMD_NULL;
+	last_command_idx = -1;
+}
+
 /**
  * ------------------------------------------------------------------------
  * Handling of repeated commands
@@ -552,8 +611,14 @@ static void cmd_set_arg(struct command *cmd, const char *name,
 
 	assert(first_empty != -1 || idx != -1);
 
-	if (idx == -1)
+	if (idx == -1) {
 		idx = first_empty;
+	} else {
+		/* Free allocated string if overwriting a string argument. */
+		if (cmd->arg[idx].type == arg_STRING) {
+			string_free((char*)(cmd->arg[idx].data.string));
+		}
+	}
 
 	cmd->arg[idx].type = type;
 	cmd->arg[idx].data = data;
@@ -752,6 +817,9 @@ void cmd_set_arg_string(struct command *cmd, const char *arg, const char *str)
 
 /**
  * Retrieve arg 'n' if a string
+ *
+ * Note that the command itself retains ownership of the pointer returned in
+ * *str:  the caller should not free that pointer.
  */
 int cmd_get_arg_string(struct command *cmd, const char *arg, const char **str)
 {
@@ -766,6 +834,9 @@ int cmd_get_arg_string(struct command *cmd, const char *arg, const char **str)
 
 /**
  * Get a string, first from the command or failing that prompt the user
+ *
+ * Note that the command itself retains ownership of the pointer returned in
+ * *str:  the caller should not free that pointer.
  */
 int cmd_get_string(struct command *cmd, const char *arg, const char **str,
 				   const char *initial, const char *title, const char *prompt)
