@@ -402,11 +402,14 @@ bool object_similar(const struct object *obj1, const struct object *obj2,
 	if (object_is_equipped(player->body, obj2))
 		return false;
 
+	/* Mimicked items do not stack */
+	if (obj1->mimicking_m_idx || obj2->mimicking_m_idx) return false;
+
 	/* If either item is unknown, do not stack */
 	if (mode & OSTACK_LIST && obj1->kind != obj1->known->kind) return false;
 	if (mode & OSTACK_LIST && obj2->kind != obj2->known->kind) return false;
 
-	/* Hack -- identical items cannot be stacked */
+	/* Identical items cannot be stacked */
 	if (obj1 == obj2) return false;
 
 	/* Require identical object kinds */
@@ -573,7 +576,8 @@ void object_origin_combine(struct object *obj1, const struct object *obj2)
  *
 * These assumptions are enforced by the "object_mergeable()" code.
  */
-static void object_absorb_merge(struct object *obj1, const struct object *obj2)
+static void object_absorb_merge(struct object *obj1, const struct object *obj2,
+		bool combine_charges_timeouts)
 {
 	int total;
 
@@ -588,14 +592,16 @@ static void object_absorb_merge(struct object *obj1, const struct object *obj2)
 	if (obj2->note)
 		obj1->note = obj2->note;
 
-	/* Combine timeouts for rod stacking */
-	if (tval_can_have_timeout(obj1))
-		obj1->timeout += obj2->timeout;
+	if (combine_charges_timeouts) {
+		/* Combine timeouts for rod stacking */
+		if (tval_can_have_timeout(obj1))
+			obj1->timeout += obj2->timeout;
 
-	/* Combine pvals for wands and staves */
-	if (tval_can_have_charges(obj1) || tval_is_money(obj1)) {
-		total = obj1->pval + obj2->pval;
-		obj1->pval = total >= MAX_PVAL ? MAX_PVAL : total;
+		/* Combine pvals for wands and staves */
+		if (tval_can_have_charges(obj1) || tval_is_money(obj1)) {
+			total = obj1->pval + obj2->pval;
+			obj1->pval = total >= MAX_PVAL ? MAX_PVAL : total;
+		}
 	}
 
 	/* Combine origin data as best we can */
@@ -657,10 +663,11 @@ void object_absorb_partial(struct object *obj1, struct object *obj2,
 		newsz2 = smallest - difference;
 	}
 
+	distribute_charges(obj2, obj1, obj2->number - newsz2, false);
 	obj1->number = newsz1;
 	obj2->number = newsz2;
 
-	object_absorb_merge(obj1, obj2);
+	object_absorb_merge(obj1, obj2, tval_is_money(obj1));
 }
 
 /**
@@ -674,7 +681,7 @@ void object_absorb(struct object *obj1, struct object *obj2)
 	/* Add together the item counts */
 	obj1->number = MIN(total, obj1->kind->base->max_stack);
 
-	object_absorb_merge(obj1, obj2);
+	object_absorb_merge(obj1, obj2, true);
 	if (known) {
 		if (!loc_is_zero(known->grid)) {
 			square_excise_object(player->cave, known->grid, known);
@@ -791,9 +798,9 @@ struct object *object_split(struct object *src, int amt)
 	assert(src->number > amt);
 
 	/* Distribute charges of wands, staves, or rods */
-	distribute_charges(src, dest, amt);
+	distribute_charges(src, dest, amt, true);
 	if (src->known)
-		distribute_charges(src->known, dest->known, amt);
+		distribute_charges(src->known, dest->known, amt, true);
 
 	/* Modify quantity */
 	dest->number = amt;
@@ -1226,7 +1233,7 @@ void push_object(struct loc grid)
 					 * Give up.  Destroy both the mimic
 					 * and the object.
 					 */
-					delete_monster_idx(obj->mimicking_m_idx);
+					delete_monster_idx(cave, obj->mimicking_m_idx);
 					if (obj->known) {
 						object_delete(player->cave, NULL, &obj->known);
 					}
